//   Copyright 2007, 2008 Nicolas Dubuit
//
//     This file is part of Immix.
//
//     Immix is free software: you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation, either version 3 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <QThread>
#include <QMutex>
#include <QList>
#include <QProgressDialog>
#include <QImage>

#ifdef Q_OS_WIN32
  #include <exiv2\exif.hpp>
  #include <exiv2\image.hpp>
#else
  #include <exiv2/exif.hpp>
  #include <exiv2/image.hpp>
#endif

#include "correlator.h"

#include "aligndialog.h"

#ifndef ALIGN_H
#define ALIGN_H

class ImmixWin ;

/*

 A note about threading.

Fist a few facts :

* To avoid blocking the interface, processing must be done in (a) separate thread(s).
* To use the full potential of multicore machines, several processing threads must be used.
* To avoid crazy memory usage, we can't use one simultaneous thread for each image to align.
Thus, each thread has to serially align several images.

Then the global design.

A thread manager dispatches work to the threads. 

Some threads are started. Each one loops, requesting one image to the manager
and subsequently working on this image.

The manager is responsible for dispatching work :
- the thread ready to work request an image number by calling itself a member of the manager.
- this member function is called from an other thread than the manager's thread itself, and as such
       * has to be very cautious
       * protects everything it modifies with mutexes
       * shouldn't emit signals or such.

The manager also collects alignment data to display. This is done fia queued signals.

*/

class AlignmentManager : public QObject
{
Q_OBJECT

    static int numThreads ;
    int numFinishedTreads ;
    QMutex numFinishedTreadsMutex ;

    AlignDialog* dialog ;
    ImmixWin* imw ;

    QMutex nextUnalignedImageMutex ;
    int nextUnalignedImage ;

    Correlator *RefWinCorrelator1 ;
    Correlator *RefWinCorrelator2 ;

public :
    AlignmentManager (ImmixWin*) ;
    ~AlignmentManager (void) ;
    int getNewImage (void) ;

public :
    QImage* getImagePtr (int) ;
    QString getImagePath (int) ;

public slots :
    void startAlign (void) ;
    void threadFinished (void) ;
    void finalProcessing (void) ;

signals :
    void allThreadsFinished (void) ;

friend class Aligner ;
friend class AligningThread ;
} ;



class AligningThread : public QThread
{
Q_OBJECT

    AlignmentManager* manager ;

public :
    AligningThread (AlignmentManager*) ;
    void run(void) ;
} ;



class Aligner : public QObject
{
Q_OBJECT

    AlignmentManager* manager ;

    Correlator* Correlator1 ;
    Correlator* Correlator2 ;
    Correlator* TargWinCorrelator1 ;
    Correlator* TargWinCorrelator2 ;

    QRect refRect1, refRect2 ;
    QImage finale ;
	QImage currentImage ;
    Exiv2::Image::AutoPtr imageExifRead ;
    Exiv2::Image::AutoPtr imageExifWrite ;

    int MkrSz, FineFactor ;

public :
    Aligner (AlignmentManager*) ;
    ~Aligner (void) ;

public slots :
    void align (int) ;

signals :
    void alignmentDone (int, double, double, double, double) ;
} ;



#endif 
